package NoRFC::Store; # $Id: Store.pm 181 2021-03-04 14:29:34Z minus $
use strict;
use utf8;

=encoding utf8

=head1 NAME

NoRFC::Store - NoRFC store class

=head1 VERSION

Version 1.02

=head1 SYNOPSIS

    use NoRFC::Store;

    my $store = NoRFC::Store->new(
        file => "/tmp/test.db",
        attributes => "RaiseError=0; PrintError=0; sqlite_unicode=1",
    );

    my $store = NoRFC::Store->new(
        dsn => "DBI:mysql:database=norfc;host=mysql.example.com",
        user => "username",
        password => "password",
        set => [
            "RaiseError        0",
            "PrintError        0",
            "mysql_enable_utf8 1",
        ],
    );

    die($store->error) unless $store->status;

=head1 DESCRIPTION

This module provides store methods.

=head2 SQLITE DDL

    CREATE TABLE "norfc" (
      "id"          INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
      "num"         CHAR(20) NOT NULL UNIQUE,
      "pub"         INTEGER NOT NULL,
      "category"    CHAR(64) DEFAULT NULL,
      "title"       TEXT DEFAULT NULL,
      "author"      CHAR(255) DEFAULT NULL,
      "abstract"    TEXT DEFAULT NULL,
      "signature"   TEXT DEFAULT NULL
    );
    CREATE UNIQUE INDEX "num" ON "norfc" (
        "num"
    );

=head2 MYSQL DDL

    CREATE DATABASE `norfc` /*!40100 DEFAULT CHARACTER SET utf8 COLLATE utf8_bin */;
    CREATE TABLE IF NOT EXISTS `norfc` (
      `id`          INT(11) NOT NULL AUTO_INCREMENT,
      `num`         VARCHAR(20) NOT NULL, -- Document number
      `pub`         INT(11) NOT NULL, -- Unixtime public date
      `category`    varchar(64) COLLATE utf8_bin NOT NULL, -- Category name
      `title`       TEXT COLLATE utf8_bin NOT NULL, -- Title
      `author`      VARCHAR(255) DEFAULT NULL,
      `abstract`    TEXT COLLATE utf8_bin DEFAULT NULL,
      `signature`   TEXT COLLATE utf8_bin DEFAULT NULL,
      PRIMARY KEY (`id`),
      UNIQUE KEY `num` (`num`)
    ) ENGINE=MyISAM DEFAULT CHARSET=utf8 COLLATE=utf8_bin;

=head2 POSTGRESQL DDL

    CREATE TABLE IF NOT EXISTS `norfc` (
      `id`          INT GENERATED BY DEFAULT AS IDENTITY PRIMARY KEY,
      `num`         CHAR(20) NOT NULL UNIQUE, -- Document number
      `pub`         INTEGER NOT NULL, -- Unixtime public date
      `category`    CHAR(64) DEFAULT NULL, -- Category name
      `title`       TEXT DEFAULT NULL, -- Title
      `author`      CHAR(255) DEFAULT NULL,
      `abstract`    TEXT COLLATE utf8_bin DEFAULT NULL,
      `signature`   TEXT COLLATE utf8_bin DEFAULT NULL
    );

=head1 METHODS

=head2 new

    my $store = NoRFC::Store->new(
        file => "/tmp/test.db",
        attributes => "RaiseError=0; PrintError=0; sqlite_unicode=1",
    );

    # ... or MySQL:

    my $store = NoRFC::Store->new(
        dsn => "DBI:mysql:database=norfc;host=mysql.example.com",
        user => "username",
        password => "password",
        set => [
            "RaiseError        0",
            "PrintError        0",
            "mysql_enable_utf8 1",
        ],
    );

Creates DBI object

=head2 add

    $store->add(
        num         => "RFC-R0000000",
        pub         => 1590000000,
        category    => "Unknown",
        title       => "Test document",
        author      => "Serz Minus",
        abstract    => "...Content of the document...",
        signature   => "***",
    ) or die($store->error);

Add new recored

=head2 count

    print $store->count();

Returns count of records

=head2 del

    $store->del("RFC-R0000000") or die($store->error);

Delete record by document number

    $store->del(1) or die($store->error);

Delete record by record id

=head2 dsn

    my $dsn = $store->dsn;

Returns DSN string of current database connection

=head2 error

    my $error = $store->error;

Returns error message

    my $status = $store->error( "Error message" );

Sets error message if argument is provided.
This method in "set" context returns status of the operation as status() method.

=head2 get

    my %data = $store->get("RFC-R0000000");

Returns data from database by document number

    my %data = $store->get(1);

Returns data from database by record id

=head2 getall

    my @table = $store->getall();
    my @table_100 = $store->getall(100); # row_count
    my @table_100 = $store->getall(100, 100); # offset, row_count

Returns data from database with limit supporting

=head2 is_sqlite

    print $store->is_sqlite ? "Is SQLite" : "Is NOT SQLite"

Returns true if type of current database is SQLite

=head2 ping

    $store->ping ? 'OK' : 'Database session is expired';

Checks the connection to database

=head2 set

    $store->set(
        num         => "RFC-R0000000",
        pub         => 1590000000,
        category    => "Unknown",
        title       => "Test document (edited)",
        author      => "Serz Minus",
        abstract    => "...Content of the document...",
        signature   => "***",
    ) or die($store->error);

Update recored by document number

    $store->set(
        id          => 1,
        num         => "RFC-R0000000",
        pub         => 1590000000,
        category    => "Unknown",
        title       => "Test document (edited)",
        author      => "Serz Minus",
        abstract    => "...Content of the document...",
        signature   => "***",
    ) or die($store->error);

Update recored by record id

=head2 status

    my $status = $store->status;
    my $status = $store->status( 1 ); # Sets the status value and returns it

Get/set BOOL status of the operation

=head2 truncate

    $store->truncate or die($store->error);

Delete all records

=head1 HISTORY

See C<Changes> file

=head1 TO DO

See C<TODO> file

=head1 BUGS

* none noted

=head1 SEE ALSO

L<CTK::DBI>, L<NoRFC>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2021 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use vars qw/$VERSION/;
$VERSION = '1.02';

use Carp;
use CTK::DBI;
use CTK::Util qw/ touch /;
use CTK::ConfGenUtil;
use CTK::TFVals qw/ :ALL /;
use File::Spec;

use constant {
    DEFAULT_DSN_MASK    => 'DBI:SQLite:dbname=%s',
    DEFAULT_DBI_ATTR    => {
            dsn         => '', # See DEFAULT_DSN_MASK
            user        => '',
            password    => '',
            set         => [
                    'RaiseError 0',
                    'PrintError 0',
                    'sqlite_unicode 1',
                ],
        },
};

use constant DDL_CREATE_TABLE => <<'DDL';
CREATE TABLE IF NOT EXISTS "norfc" (
    "id"        INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
    "num"       CHAR(20) NOT NULL UNIQUE,
    "pub"       INTEGER NOT NULL,
    "category"  CHAR(64) DEFAULT NULL,
    "title"     TEXT DEFAULT NULL,
    "author"    CHAR(255) DEFAULT NULL,
    "abstract"  TEXT DEFAULT NULL,
    "signature" TEXT DEFAULT NULL
)
DDL
use constant DDL_CREATE_INDEX => <<'DDL';
CREATE UNIQUE INDEX "num" ON "norfc" (
    "num"
)
DDL

# INSERT INTO norfc
#    (`num`,`pub`,`category`,`title`,`author`,`abstract`,`signature`)
# VALUES
#    ("RFC-R0000000", 1590000000, "Unknown", "Test document", "Serz Minus", "...Content of the document...", "***")
use constant DML_ADD => <<'DML';
INSERT INTO `norfc`
    (`num`,`pub`,`category`,`title`,`author`,`abstract`,`signature`)
VALUES
    (?,?,?,?,?,?,?)
DML

use constant DML_GET_COUNT => <<'DML';
SELECT COUNT(`id`) AS `cnt`
FROM `norfc`
DML

use constant DML_GET_BY_ID => <<'DML';
SELECT `id`,`num`,`pub`,`category`,`title`,`author`,`abstract`,`signature`
FROM `norfc`
WHERE `id` = ?
DML

use constant DML_GET_BY_NUM => <<'DML';
SELECT `id`,`num`,`pub`,`category`,`title`,`author`,`abstract`,`signature`
FROM `norfc`
WHERE `num` = ?
DML

use constant DML_GETALL => <<'DML';
SELECT `id`,`num`,`pub`,`category`,`title`,`author`,`abstract`,`signature`
FROM `norfc`
ORDER BY `pub` ASC
DML

use constant DML_DEL_BY_ID => <<'DML';
DELETE FROM `norfc` WHERE `id` = ?
DML

use constant DML_DEL_BY_NUM => <<'DML';
DELETE FROM `norfc` WHERE `num` = ?
DML

use constant DML_TRUNCATE => <<'DML';
DELETE FROM `norfc`
DML

use constant DML_SET_BY_ID => <<'DML';
UPDATE `norfc`
SET `num` = ?, `pub` = ?, `category` = ?, `title` = ?, `author` = ?, `abstract` = ?, `signature` = ?
WHERE `id` = ?
DML

use constant DML_SET_BY_NUM => <<'DML';
UPDATE `norfc`
SET `num` = ?, `pub` = ?, `category` = ?, `title` = ?, `author` = ?, `abstract` = ?, `signature` = ?
WHERE `num` = ?
DML

sub new {
    my $class = shift;
    my %args = @_;
    my $args_set = $args{'set'};
    unless ($args{dsn}) {
        my $dda = DEFAULT_DBI_ATTR;
        foreach (keys %$dda) {
            next if $_ eq 'set';
            $args{$_} //= $dda->{$_}
        }
    }
    my $file = $args{file} // "";
    my $dsn = $args{dsn} || sprintf(DEFAULT_DSN_MASK, $file);
    my %attrs_from_set = _set2attr($args_set);
    my %attrs_from_str = $args{attributes} ? _parseDBAttributes($args{attributes}) : ();
    my %attrs = (%attrs_from_set, %attrs_from_str);
       %attrs = _set2attr(DEFAULT_DBI_ATTR()->{set}) unless %attrs;

    # DB
    my $db = CTK::DBI->new(
        -dsn        => $dsn,
        -debug      => 0,
        -username   => $args{'user'},
        -password   => $args{'password'},
        -attr       => {%attrs},
        $args{timeout} ? (
            -timeout_connect => $args{timeout},
            -timeout_request => $args{timeout},
        ) : (),
    );
    my $dbh = $db->connect if $db;

    # SQLite
    my $fnew = 0;
    my $issqlite = 0;
    if ($dbh && $dsn =~ /SQLite/i) { # If SQLite (default)
        $file = $dbh->sqlite_db_filename();
        unless ($file && (-e $file) && !(-z $file)) {
            touch($file);
            chmod(0666, $file);
            $fnew = 1;
        }
        $issqlite = 1;
    }

    # Defaults
    my $status = 1; # Ok
    my $error = "";
    if (!$db) {
        $error = sprintf("Can't init database \"%s\"", $dsn);
        $status = 0;
    } elsif (!$dbh) {
        $error = sprintf("Can't connect to database \"%s\": %s", $dsn, $DBI::errstr || "unknown error");
        $status = 0;
    } elsif ($fnew) {
        $db->execute(DDL_CREATE_TABLE);
        $db->execute(DDL_CREATE_INDEX) unless $dbh->err;
        $error = $dbh->errstr();
        $status = 0 if $dbh->err;
    }
    unless ($error) { # No errors
        unless ($dbh->ping) {
            $error = sprintf("Can't init database \"%s\". Ping failed: %s",
                $dsn, $dbh->errstr() || "unknown error");
            $status = 0;
        }
    }

    my $self = bless {
            file    => $file,
            issqlite=> $issqlite,
            dsn     => $dsn,
            error   => $error // "",
            dbi     => $db,
            status  => $status,
        }, $class;

    return $self;
}

sub status {
    my $self = shift;
    my $value = shift;
    return fv2zero($self->{status}) unless defined($value);
    $self->{status} = $value ? 1 : 0;
    return $self->{status};
}
sub error {
    my $self = shift;
    my $value = shift;
    return uv2null($self->{error}) unless defined($value);
    $self->{error} = $value;
    $self->status($value ne "" ? 0 : 1);
    return $value;
}
sub ping {
    my $self = shift;
    return 0 unless $self->{dsn};
    my $dbi = $self->{dbi};
    return 0 unless $dbi;
    my $dbh = $dbi->{dbh};
    return 0 unless $dbh;
    return 0 unless $dbh->can('ping');
    return $dbh->ping();
}
sub dsn {
    my $self = shift;
    return $self->{dsn};
}
sub is_sqlite {
    my $self = shift;
    return $self->{issqlite} ? 1 : 0;
}

# CRUD Methods

sub add {
    my $self = shift;
    my %data = @_;
    return unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    # Add
    $dbi->execute(DML_ADD,
        $data{num},
        $data{pub} || time(),
        $data{category},
        $data{title},
        $data{author},
        $data{abstract},
        $data{signature},
    );
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't insert new record: %s", uv2null($dbi->connect->errstr)));
        return;
    }

    return 1;
}
sub count {
    my $self = shift;
    return 0 unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    my $cnt = $dbi->field(DML_GET_COUNT) || 0;
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't get count: %s", uv2null($dbi->connect->errstr)));
        return 0;
    }

    return $cnt;
}
sub get {
    my $self = shift;
    my $id = shift || 0;
    return () unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    my %rec = $dbi->recordh(is_int($id) ? DML_GET_BY_ID : DML_GET_BY_NUM, $id);
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't get record: %s", uv2null($dbi->connect->errstr)));
        return ();
    }

    return %rec;
}
sub getall {
    my $self = shift;
    my $row_count = pop;
    my $offset = pop;
    return () unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};
    my $limit = $row_count
        ? $offset
            ? sprintf(" LIMIT %d, %d", $offset, $row_count)
            : sprintf(" LIMIT %d", $row_count)
        : "";

    my @tbl = $dbi->table(sprintf("%s%s", DML_GETALL,  $limit ));
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't get records: %s", uv2null($dbi->connect->errstr)));
        return ();
    }
    return @tbl;
}
sub del {
    my $self = shift;
    my $id = shift || 0;
    return 0 unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    $dbi->execute(is_int($id) ? DML_DEL_BY_ID : DML_DEL_BY_NUM, $id);
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't delete record: %s", uv2null($dbi->connect->errstr)));
        return 0;
    }
    return 1;
}
sub truncate {
    my $self = shift;
    return 0 unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    $dbi->execute(DML_TRUNCATE);
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't truncate table: %s", uv2null($dbi->connect->errstr)));
        return 0;
    }
    return 1;
}
sub set {
    my $self = shift;
    my %data = @_;
    return unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    # Set by id or num
    $dbi->execute($data{id} ? DML_SET_BY_ID : DML_SET_BY_NUM,
        $data{num},
        $data{pub} || time(),
        $data{category},
        $data{title},
        $data{author},
        $data{abstract},
        $data{signature},
        $data{id} || $data{num},
    );
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't insert new record: %s", uv2null($dbi->connect->errstr)));
        return;
    }

    return 1;
}

sub _set2attr {
    my $in = shift;
    my $attr = is_array($in) ? $in : array($in => "set");
    my %attrs;
    foreach (@$attr) {
        $attrs{$1} = $2 if $_ =~ /^\s*(\S+)\s+(.+)$/;
    }
    return %attrs;
}
sub _parseDBAttributes { # Example: foo=123, bar=gii,baz;qux=werwe
    my $t = shift || return ();
    my %as = ();
    for ($t ? (split /\s*[,;]\s*/, $t) : ()) {
        my ($k,$v) = (split /\s*=>?\s*|\s*\:\s*/, $_);
        $as{$k} = $v;
    }
    return %as;
}

1;

__END__
